# """Console script for fintorch."""

from typing import Any
import click
import optuna
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.callbacks.early_stopping import EarlyStopping
from lightning.pytorch.cli import LightningCLI
from optuna.integration import WeightsAndBiasesCallback

from fintorch.datasets import elliptic as e
from fintorch.datasets import ellipticpp as epp
from fintorch.models.graph.graphbean.wand.sweep import objective


@click.group()
def fintorch() -> None:
    """FinTorch CLI - Your financial AI toolkit"""
    pass


@fintorch.command()
@click.argument("dataset")
def datasets(dataset: str) -> None:
    """Download financial datasets"""
    # Implement your dataset download logic here
    click.echo(f"Downloading dataset: {dataset}")
    if dataset == "elliptic":
        e.TransactionDataset("~/.fintorch_data", force_reload=True)
    elif dataset == "ellipticpp":
        epp.TransactionActorDataset("~/.fintorch_data", force_reload=True)


@fintorch.command()
@click.option("--model", required=True, help="Name of the model to use.")
@click.option("--predict", required=True, help="Type of prediction to perform.")
@click.option("--max_epochs", required=False, help="Max epochs")
def sweep(model: str, predict: Any, max_epochs: int) -> None:
    """Sweep your financial models"""
    # Implement your model training logic here
    click.echo(f"Starting sweep for model: {model}")

    if model == "graphbean_elliptic":
        wandb_kwargs = {"project": f"graphbean-{predict}"}
        wandbc = WeightsAndBiasesCallback(wandb_kwargs=wandb_kwargs, as_multirun=True)

        study = optuna.create_study(
            direction="maximize", pruner=optuna.pruners.MedianPruner()
        )

        def wrapped_objective(trial: Any) -> float:
            return objective(trial, max_epochs, predict)

        # study.optimize(objective, n_trials=100)
        study.optimize(wrapped_objective, n_trials=100, callbacks=[wandbc])

        print("Number of finished trials: {}".format(len(study.trials)))

        print("Best trial:")
        trial = study.best_trial

        print("  Value: {}".format(trial.value))

        print("  Params: ")
        for key, value in trial.params.items():
            print("    {}: {}".format(key, value))


def trainer() -> None:
    cli = LightningCLI(
        run=False,
        save_config_callback=None,
        trainer_defaults={
            "max_epochs": 10,
            "callbacks": [
                ModelCheckpoint(monitor="val_loss", mode="min"),
                EarlyStopping(monitor="train_loss", mode="min"),
            ],
        },
    )

    cli.trainer.fit(cli.model, datamodule=cli.datamodule)



import os
import tempfile

import pytest

from fintorch.datasets.elliptic import TransactionDataset


@pytest.fixture
def dataset():
    root = tempfile.mkdtemp()
    return TransactionDataset(root, force_reload=True)


def test_raw_file_names(dataset):
    expected_files = [
        "elliptic_bitcoin_dataset/elliptic_txs_features.csv",
        "elliptic_bitcoin_dataset/elliptic_txs_edgelist.csv",
        "elliptic_bitcoin_dataset/elliptic_txs_classes.csv",
    ]
    assert dataset.raw_file_names == expected_files


def test_processed_file_names(dataset):
    expected_files = ["transaction_graph_v1.pt"]
    assert dataset.processed_file_names == expected_files


def test_download(dataset):
    dataset.download()
    assert os.path.exists(dataset.raw_dir)


def test_process(dataset):
    dataset.process()
    assert os.path.exists(dataset.processed_paths[0])


def test_data_loading(dataset):
    data = dataset[0]
    assert "x" in data
    assert "edge_index" in data
    assert "y" in data
    assert "train_mask" in data
    assert "val_mask" in data
    assert "test_mask" in data



import logging
import os
import shutil
from datetime import datetime
from typing import Any, Generator, List
from zipfile import ZipFile

import polars as pol
from torch.utils.data import IterableDataset


class MarketDataset(IterableDataset):  # type: ignore
    """PyTorch IterableDataset for Jane Street Market Data.

    Args:
        root (str): Root directory for dataset storage
        split (str, optional): Data split ('train' or 'test'). Defaults to "train"
        force_reload (bool, optional): Force data reprocessing. Defaults to False
        batch_size (int, optional): Number of samples per batch. Defaults to 10000000
    """

    REQUIRED_COLUMNS = {
        "train": {"date_id", "time_id", "symbol_id", "responder_6"},
        "test": {"date_id", "time_id", "symbol_id"},
    }

    def __init__(
        self,
        root: str,
        split: str = "train",
        force_reload: bool = False,
        batch_size: int = 100000,
    ):
        super().__init__()
        if split not in ["train", "test"]:
            raise ValueError("split must be either 'train' or 'test'")

        self.root = os.path.expanduser(root)
        self.batch_size = batch_size
        self.offset = 0
        self.split = split  # 'train' or 'test'

        logging.info(f"Load {self.split} market data")
        self.setup_directories()

        if force_reload or not all(os.path.exists(path) for path in self.raw_paths()):
            # If we want to force reload, or a processed file is missing, start the processing
            self.download()  # Download auction data
            self.process()

        if (
            force_reload
            or not os.path.exists(os.path.join(self.root, "processed", self.split))
            or len(os.listdir(os.path.join(self.root, "processed", self.split))) == 0
        ):
            self.process()

        self.load()

    def __iter__(self) -> Generator[tuple[int, pol.DataFrame], Any, None]:
        self.offset = 0
        idx = 0
        while True:
            # Slice the LazyFrame to get the next batch
            batch_df = self.data.slice(self.offset, self.batch_size).collect()

            # If the batch is empty, we've reached the end of the dataset
            if batch_df.is_empty():
                break

            # Yield the batch data
            yield idx, batch_df

            # Update the offset for the next batch
            self.offset += self.batch_size
            idx += 1

    def raw_paths(self) -> List[str]:
        return [
            os.path.join(self.root, path)
            for path in [
                "raw/features.csv",
                "raw/responders.csv",
                "raw/sample_submission.csv",
                "raw/train.parquet",
                "raw/test.parquet",
                "raw/lags.parquet",
            ]
        ]

    def batch_raw(
        self, raw_data: pol.LazyFrame
    ) -> Generator[tuple[int, pol.DataFrame], Any, None]:
        idx = 0
        while True:
            # Slice the LazyFrame to get the next batch
            batch_df = raw_data.slice(self.offset, self.batch_size).collect()

            # If the batch is empty, we've reached the end of the dataset
            if batch_df.is_empty():
                break

            # Yield the batch data
            yield idx, batch_df

            # Update the offset for the next batch
            self.offset += self.batch_size
            idx += 1

    def process(self) -> None:
        logging.info(f"Processing: apply transformations to {self.split} market data")

        try:
            path = os.path.join(self.root, "raw", f"{self.split}.parquet")
            raw = pol.scan_parquet(path)
        except Exception as e:
            raise RuntimeError(f"Failed to read parquet file {path}: {str(e)}") from e

        processed_dir = os.path.join(self.root, "processed", self.split)
        if os.path.exists(processed_dir) and len(os.listdir(processed_dir)) > 0:
            # Clean up old files
            try:
                logging.info("Cleaning up old processed files")
                shutil.rmtree(processed_dir)
                os.mkdir(processed_dir)
            except Exception as e:
                logging.error(f"An error occurred cleaning up the files: {e}")
                raise RuntimeError(
                    f"Failed to clean up processed directory: {str(e)}"
                ) from e

        for idx, batch in self.batch_raw(raw):
            try:
                batch_prep = self.preprocess_batch(batch)
                batch_prep = batch_prep.with_columns(pol.lit(idx).alias("partition_id"))
                batch_prep.write_parquet(
                    processed_dir,
                    partition_by=["partition_id", "unique_id"],
                )
                del batch_prep
            except Exception as e:
                raise RuntimeError(f"Failed to process batch {idx}: {str(e)}") from e

    def download(self) -> None:
        from kaggle.api.kaggle_api_extended import KaggleApi  # type: ignore

        logging.info("Downloading dataset from Kaggle")
        try:
            api = KaggleApi()
            api.read_config_file()
            api.authenticate()
        except Exception as e:
            raise RuntimeError(
                "Failed to authenticate with Kaggle API. Ensure you have a valid kaggle.json file. See https://github.com/Kaggle/kaggle-api/blob/main/docs/README.md for more info."
            ) from e

        try:
            competition = "jane-street-real-time-market-data-forecasting"
            api.competition_download_files(
                competition, path=os.path.join(self.root, "raw"), quiet=False
            )
        except Exception as e:
            raise RuntimeError(f"Failed to download compitition files:{str(e)}") from e

        path = os.path.join(self.root, "raw", competition + ".zip")

        try:
            with ZipFile(path) as zObject:
                zObject.extractall(os.path.join(self.root, "raw"))

            os.remove(path)
        except Exception as e:
            raise RuntimeError(f"Failed to extract dataset: {str(e)}") from e

        # Verify all required files are present
        missing_files = [p for p in self.raw_paths() if not os.path.exists(p)]
        if missing_files:
            raise RuntimeError(
                f"Missing required files after download: {missing_files}"
            )

    def load(self) -> None:
        path = os.path.join(self.root, "processed", self.split)
        self.data = pol.scan_parquet(path)

    def preprocess_batch(self, batch: pol.DataFrame) -> pol.DataFrame:
        # Validate required columns
        missing_cols = self.REQUIRED_COLUMNS[self.split] - set(batch.columns)
        if missing_cols:
            raise ValueError(f"Missing required columns: {missing_cols}")

        df = (
            self.map_to_datetime(batch)
            .fill_nan(0)
            .fill_null(0)
            .rename({"symbol_id": "unique_id"})
        )
        if self.split == "train":
            df = df.rename({"responder_6": "y"})
        return df

    @staticmethod
    def map_to_datetime(df: pol.DataFrame) -> pol.DataFrame:
        start_date = datetime(2023, 1, 1)
        return df.with_columns(
            (
                pol.lit(start_date)
                + (pol.col("date_id") * pol.duration(minutes=9))
                + pol.duration(hours=12)
                + (pol.col("time_id") * pol.duration(seconds=1))
            ).alias("ds")
        )

    def setup_directories(self) -> None:
        try:
            os.makedirs(self.root, exist_ok=True)
            os.makedirs(os.path.join(self.root, "raw"), exist_ok=True)
            os.makedirs(os.path.join(self.root, "processed"), exist_ok=True)
        except OSError as e:
            logging.error(f"Failed to create directories: {str(e)}")
            raise RuntimeError(f"Failed to set up directories: {str(e)}") from e



import logging
from pathlib import Path

from fintorch.datasets.auctiondata import AuctionDataset

logging.basicConfig(level=logging.INFO)

data_path = Path("~/.fintorch_data/auctiondata-optiver/").expanduser()
auction_data = AuctionDataset(data_path, False)

print(f"Length of the dataset:{len(auction_data)} \n Print first 10 records:")

for i in range(10):
    print(auction_data[i])



import polars as pl
import torch
from torch_geometric.loader import LinkNeighborLoader

from fintorch.datasets.ellipticpp import EllipticppDataModule, TransactionActorDataset


def test_download():
    dataset = TransactionActorDataset(root="test_data", force_reload=True)
    dataset.download()
    assert len(dataset.downloaded_files) == len(dataset.raw_file_names)


def test_map_classes():
    dataset = TransactionActorDataset(root="test_data")
    df = pl.DataFrame({"class": ["3", "2", "1"]})
    mapped_df = dataset.map_classes(df)
    assert mapped_df["class"].to_list() == [2, 1, 0]


def test_split_data():
    dataset = TransactionActorDataset(root="test_data")
    num_data = 100
    train_mask, val_mask, test_mask = dataset.split_data(num_data, splits=[0.8, 0.1])
    assert train_mask.sum() == 80
    assert val_mask.sum() == 10
    assert test_mask.sum() == 10


def test_prepare_edge_index():
    dataset = TransactionActorDataset(root="test_data")
    edgelist = pl.DataFrame(
        {"input_address": ["a", "b", "c"], "output_address": ["x", "y", "z"]}
    )
    mapping_dict = {
        "input_address": {"a": 0, "b": 1, "c": 2},
        "output_address": {"x": 3, "y": 4, "z": 5},
    }
    edge_index = dataset.prepare_edge_index(edgelist, mapping_dict)
    assert torch.all(torch.eq(edge_index, torch.tensor([[0, 1, 2], [3, 4, 5]])))


def test_process():
    dataset = TransactionActorDataset(root="test_data")
    dataset.process()
    dataset = dataset[0]

    assert dataset["wallets"].x.shape[0] == dataset["wallets"].y.shape[0]
    assert dataset["wallets"].train_mask.shape[0] == dataset["wallets"].y.shape[0]
    assert dataset["wallets"].val_mask.shape[0] == dataset["wallets"].y.shape[0]
    assert dataset["wallets"].test_mask.shape[0] == dataset["wallets"].y.shape[0]
    assert (
        dataset["wallets"].train_mask.sum()
        + dataset["wallets"].val_mask.sum()
        + dataset["wallets"].test_mask.sum()
        == dataset["wallets"].y.shape[0]
    )

    assert dataset["transactions"].x.shape[0] == dataset["transactions"].y.shape[0]
    assert (
        dataset["transactions"].train_mask.shape[0]
        == dataset["transactions"].y.shape[0]
    )
    assert (
        dataset["transactions"].val_mask.shape[0] == dataset["transactions"].y.shape[0]
    )
    assert (
        dataset["transactions"].test_mask.shape[0] == dataset["transactions"].y.shape[0]
    )
    assert (
        dataset["transactions"].train_mask.sum()
        + dataset["transactions"].val_mask.sum()
        + dataset["transactions"].test_mask.sum()
        == dataset["transactions"].y.shape[0]
    )

    assert dataset["wallets", "to", "transactions"]["edge_index"].shape[0] == 2
    assert dataset["transactions", "to", "wallets"]["edge_index"].shape[0] == 2
    assert dataset["wallets", "to", "wallets"]["edge_index"].shape[0] == 2
    assert dataset["transactions", "to", "transactions"]["edge_index"].shape[0] == 2


def test_setup():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    assert data_module.train_data is not None
    assert data_module.val_data is not None
    assert data_module.test_data is not None


def test_train_dataloader():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    train_dataloader = data_module.train_dataloader()
    assert train_dataloader is not None


def test_train_dataloader_type():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    train_dataloader = data_module.train_dataloader()
    assert isinstance(train_dataloader, LinkNeighborLoader)


def test_val_dataloader_type():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    val_dataloader = data_module.val_dataloader()
    assert isinstance(val_dataloader, LinkNeighborLoader)


def test_val_dataloader():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    val_dataloader = data_module.val_dataloader()
    assert val_dataloader is not None


def test_link_neighbor_loader_type():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    train_dataloader = data_module.train_dataloader()
    assert isinstance(train_dataloader, LinkNeighborLoader)


def test_test_dataloader():
    data_module = EllipticppDataModule(edge=("wallets", "to", "transactions"))
    data_module.setup()
    test_dataloader = data_module.test_dataloader()
    assert test_dataloader is not None


test_download()
test_map_classes()
test_split_data()
test_prepare_edge_index()
test_process()
test_setup()
test_train_dataloader()
test_val_dataloader()
test_test_dataloader()



import tempfile
from datetime import date
from pathlib import Path
from unittest.mock import patch

import pytest

from fintorch.datasets.stockticker import StockTicker


def create_mock_stock_data():
    import pandas as pd

    data = pd.read_pickle("tests/datasets/testsdata.pickle")
    return data


@patch("fintorch.datasets.stockticker.yf.download")
def test_download(mock_download):
    # Create temporary directory for test data
    with tempfile.TemporaryDirectory() as tmp_dir:
        data_path = Path(tmp_dir)
        tickers = ["AAPL", "MSFT", "GOOG"]
        start_date = date(2015, 1, 1)
        end_date = date(2023, 6, 30)
        ticker_index = {ticker: index for index, ticker in enumerate(tickers)}

        mock_download.return_value = create_mock_stock_data()

        dataset = StockTicker(
            data_path,
            tickers=tickers,
            start_date=start_date,
            end_date=end_date,
            mapping=ticker_index,
            force_reload=True,
        )

    assert dataset.df_timeseries_dataset.count().shape == (1, 3)
    ds, y, unique_id = dataset.df_timeseries_dataset.count()
    assert ds[0] > 0
    assert y[0] > 0
    assert unique_id[0] > 0
    assert ds[0] == y[0]
    assert y[0] == unique_id[0]


@patch("fintorch.datasets.stockticker.yf.download")
def test_no_tickers(mock_download):
    # Create temporary directory for test data
    with tempfile.TemporaryDirectory() as tmp_dir:
        data_path = Path(tmp_dir)
        tickers = "AAPL"
        start_date = date(2015, 1, 1)
        end_date = date(2023, 6, 30)
        ticker_index = {ticker: index for index, ticker in enumerate(tickers)}

        mock_download.return_value = create_mock_stock_data()

        with pytest.raises(AssertionError) as exc_info:
            StockTicker(
                data_path,
                tickers=tickers,
                start_date=start_date,
                end_date=end_date,
                mapping=ticker_index,
                force_reload=True,
            )
        assert str(exc_info.value) == "tickers must be a list"


@patch("fintorch.datasets.stockticker.yf.download")
def test_wrong_date(mock_download):
    # Create temporary directory for test data
    with tempfile.TemporaryDirectory() as tmp_dir:
        data_path = Path(tmp_dir)
        tickers = ["AAPL"]
        end_date = date(2015, 1, 1)
        start_date = date(2023, 6, 30)
        ticker_index = {ticker: index for index, ticker in enumerate(tickers)}
        mock_download.return_value = create_mock_stock_data()
        with pytest.raises(ValueError) as exc_info:
            StockTicker(
                data_path,
                tickers=tickers,
                start_date=start_date,
                end_date=end_date,
                mapping=ticker_index,
                force_reload=True,
            )
        assert str(exc_info.value) == "start_date must be before end_date."


@patch("fintorch.datasets.stockticker.yf.download")
def test_wrong_date_type(mock_download):
    # Create temporary directory for test data
    with tempfile.TemporaryDirectory() as tmp_dir:
        data_path = Path(tmp_dir)
        tickers = ["AAPL"]
        end_date = date(2015, 1, 1)
        start_date = "wrong"
        ticker_index = {ticker: index for index, ticker in enumerate(tickers)}
        mock_download.return_value = create_mock_stock_data()
        with pytest.raises(AssertionError) as exc_info:
            StockTicker(
                data_path,
                tickers=tickers,
                start_date=start_date,
                end_date=end_date,
                mapping=ticker_index,
                force_reload=True,
            )
        assert str(exc_info.value) == "start_date must be a Date object"



#!/usr/bin/env python
import os

PROJECT_DIRECTORY = os.path.realpath(os.path.curdir)


def remove_file(filepath):
    os.remove(os.path.join(PROJECT_DIRECTORY, filepath))


if __name__ == "__main__":
    if "{{ cookiecutter.create_author_file }}" != "y":
        remove_file("AUTHORS.rst")
        remove_file("docs/authors.rst")

    if "no" in "{{ cookiecutter.command_line_interface|lower }}":
        cli_file = os.path.join("{{ cookiecutter.project_slug }}", "cli.py")
        remove_file(cli_file)

    if "Not open source" == "{{ cookiecutter.open_source_license }}":
        remove_file("LICENSE")



from datetime import date
from pathlib import Path

import torch
from neuralforecast import NeuralForecast
from neuralforecast.auto import AutoNBEATS, AutoNHITS, AutoTFT
from neuralforecast.losses.numpy import mae, mse
from neuralforecast.losses.pytorch import MAE, HuberLoss
from ray import tune

from fintorch.datasets import stockticker

torch.set_float32_matmul_precision("medium")

tickers = ["AAPL", "MSFT", "GOOG"]
data_path = Path("~/.fintorch_data/stocktickers/").expanduser()
start_date = date(2015, 1, 1)
end_date = date(2023, 6, 30)

# Create a dictionary mapping from tickers to index
ticker_index = {ticker: index for index, ticker in enumerate(tickers)}

# Auto model settings
horizon = 96
num_samples = 10
freq = "1d"


def main():
    max_steps = tune.choice([500, 1500, 2000])

    tft_config = {
        "input_size": tune.choice([10, 20, 30]),
        "hidden_size": tune.choice([64, 128, 256]),
        "n_head": tune.choice([4, 8]),
        "learning_rate": tune.loguniform(1e-4, 1e-1),
        "scaler_type": tune.choice([None, "robust", "standard"]),
        "max_steps": max_steps,
        "batch_size": tune.choice([32, 64, 128, 256]),
        "windows_batch_size": tune.choice([128, 256, 512, 1024]),
        "random_seed": tune.randint(1, 20),
    }

    nbeats_config = {
        "input_size": tune.choice([10, 20, 30]),
        "n_harmonics": tune.choice([2, 3, 4]),
        "n_polynomials": tune.choice([1, 2, 3]),
        "stack_types": tune.choice([["trend", "seasonality"]]),
        "n_blocks": tune.choice([[3, 3], [4, 4], [5, 5]]),
        "mlp_units": tune.choice(
            [
                [[1024, 2048], [2048, 2048]],
                [[2048, 4096], [4096, 4096]],
                [[2048, 4096], [4096, 8192]],
            ]
        ),
        "dropout_prob_theta": 0.0,
        "activation": tune.choice(["ReLU", "LeakyReLU"]),
        "shared_weights": tune.choice([True, False]),
        "valid_loss": None,
        "max_steps": max_steps,
        "learning_rate": tune.loguniform(1e-5, 1e-2),
        "num_lr_decays": tune.choice([2, 3, 4]),
        "early_stop_patience_steps": -1,
        "val_check_steps": 100,
        "batch_size": 1024,
        "valid_batch_size": None,
        "windows_batch_size": 1024,
        "inference_windows_batch_size": -1,
        "start_padding_enabled": False,
        "step_size": 1,
        "scaler_type": "identity",
        "random_seed": 1,
        "num_workers_loader": 0,
    }

    nhits_config = {
        # "learning_rate": tune.choice([1e-3]),                                     # Initial Learning rate
        "learning_rate": tune.loguniform(1e-5, 5e-3),
        "max_steps": max_steps,  # Number of SGD steps
        "input_size": tune.choice([7 * horizon]),  # input_size = multiplier * horizon
        "batch_size": tune.choice([7]),  # Number of series in windows
        "windows_batch_size": tune.choice([256]),  # Number of windows in batch
        "n_pool_kernel_size": tune.choice(
            [[2, 2, 2], [16, 8, 1]]
        ),  # MaxPool's Kernelsize
        "n_freq_downsample": tune.choice(
            [[(96 * 7) // 2, 96 // 2, 1], [(24 * 7) // 2, 24 // 2, 1], [1, 1, 1]]
        ),  # Interpolation expressivity ratios
        "dropout_prob_theta": tune.choice([0.5]),  # Dropout regularization
        "activation": tune.choice(["ReLU"]),  # Type of non-linear activation
        "n_blocks": tune.choice([[1, 1, 1]]),  # Blocks per each 3 stacks
        "mlp_units": tune.choice(
            [[[512, 512], [512, 512], [512, 512]]]
        ),  # 2 512-Layers per block for each stack
        "interpolation_mode": tune.choice(
            ["linear"]
        ),  # Type of multi-step interpolation
        "val_check_steps": tune.choice([100]),  # Compute validation every 100 epochs
        "random_seed": tune.randint(1, 10),
    }

    models = [
        AutoTFT(
            h=horizon,
            loss=HuberLoss(delta=0.5),
            valid_loss=MAE(),
            config=tft_config,
            num_samples=num_samples,
            refit_with_val=True,
        ),
        AutoNBEATS(
            h=horizon,
            loss=HuberLoss(delta=0.5),
            valid_loss=MAE(),
            config=nbeats_config,
            num_samples=num_samples,
            refit_with_val=True,
        ),
        AutoNHITS(
            h=horizon,
            loss=HuberLoss(delta=0.5),
            valid_loss=MAE(),
            config=nhits_config,
            num_samples=num_samples,
            refit_with_val=True,
        ),
    ]

    stockdata = stockticker.StockTicker(
        data_path,
        tickers=tickers,
        start_date=start_date,
        end_date=end_date,
        mapping=ticker_index,
        force_reload=True,
    )

    n_time = len(stockdata.df_timeseries_dataset["ds"].unique())
    val_size = int(n_time * 0.1)
    test_size = int(n_time * 0.1)

    nf = NeuralForecast(models=models, freq=freq)
    Y_hat_df = nf.cross_validation(
        df=stockdata.df_timeseries_dataset,
        val_size=val_size,
        test_size=test_size,
        n_windows=None,
    ).to_pandas()

    # List of model names
    model_names = ["AutoTFT", "AutoNHITS", "AutoNBEATS"]

    # Number of unique series
    n_series = len(stockdata.df_timeseries_dataset["unique_id"].unique())

    # Iterate over each model to compute and print MSE and MAE
    for model_name in model_names:
        # Extract true values and predictions
        y_true = Y_hat_df.y.values
        y_hat = Y_hat_df[model_name].values

        # Reshape to (n_series, n_windows, horizon)
        y_true = y_true.reshape(n_series, -1, horizon)
        y_hat = y_hat.reshape(n_series, -1, horizon)

        # Compute MSE and MAE
        mse_score = mse(y_hat, y_true)
        mae_score = mae(y_hat, y_true)

        # Print results
        print(f"\nModel: {model_name}")
        print(f"MSE: {mse_score}")
        print(f"MAE: {mae_score}")


if __name__ == "__main__":
    main()



# from fintorch.datasets import elliptic
import lightning.pytorch as pl
import torch.nn as nn
import torch.optim as optim
import torch_geometric.nn as geom_nn
import torchmetrics
from torch import Tensor
from torch_geometric.loader import DataLoader

from fintorch.datasets import elliptic

# Load the elliptic dataset
elliptic_dataset = elliptic.TransactionDataset("~/.fintorch_data", force_reload=True)


class GNNModel(nn.Module):
    def __init__(
        self,
        c_in: int,
        c_hidden: int,
        c_out: int,
        num_layers: int = 5,
        dp_rate: float = 0.1,
        **kwargs,
    ):
        """
        Initialize the Elliptical class.

        Args:
            c_in (int): Number of input channels.
            c_hidden (int): Number of hidden channels.
            c_out (int): Number of output channels.
            num_layers (int, optional): Number of GNN layers. Defaults to 5.
            dp_rate (float, optional): Dropout rate. Defaults to 0.1.
            **kwargs: Additional keyword arguments to be passed to the GNN layers.

        Returns:
            None
        """

        super().__init__()
        gnn_layer = geom_nn.GCNConv

        layers = []
        in_channels, out_channels = c_in, c_hidden
        for _ in range(num_layers - 1):
            layers += [
                gnn_layer(in_channels=in_channels, out_channels=out_channels, **kwargs),
                nn.ReLU(inplace=True),
                nn.Dropout(dp_rate),
            ]
            in_channels = c_hidden
        layers += [gnn_layer(in_channels=in_channels, out_channels=c_out, **kwargs)]
        self.layers = nn.ModuleList(layers)

    def forward(self, x: Tensor, edge_index: Tensor) -> Tensor:
        """
        Forward pass of the model.

        Args:
            x (Tensor): Input tensor.
            edge_index (Tensor): Edge index tensor.

        Returns:
            Tensor: Output tensor.
        """

        for layer in self.layers:
            if isinstance(layer, geom_nn.MessagePassing):
                # In case of a geom layer, also pass the edge_index list
                x = layer(x, edge_index)
            else:
                x = layer(x)

        return x


class GNN(pl.LightningModule):
    def __init__(self, **model_kwargs):
        super().__init__()
        self.save_hyperparameters()

        self.accuracy = torchmetrics.classification.Accuracy(
            task="multiclass", num_classes=3
        )

        self.model = GNNModel(**model_kwargs)
        self.loss_module = nn.CrossEntropyLoss()

    def forward(self, data, mode="train"):
        x, edge_index = data.x, data.edge_index
        x = self.model(x, edge_index)

        # Get the mask
        if mode == "train":
            mask = data.train_mask
        elif mode == "val":
            mask = data.val_mask
        elif mode == "test":
            mask = data.test_mask
        else:
            assert False, f"Unknown forward mode: {mode}"

        # Calculate the loss for the mask
        loss = self.loss_module(x[mask], data.y[mask].long())
        pred = x[mask].argmax(dim=-1)

        return loss, pred, data.y[mask]

    def configure_optimizers(self):
        optimizer = optim.Adam(self.parameters(), lr=0.05)
        return optimizer

    def training_step(self, batch, batch_idx):
        loss, preds, y = self.forward(batch, mode="train")

        # log step metric
        self.accuracy(preds, y)
        self.log("train_acc_step", self.accuracy)

        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, preds, y = self.forward(batch, mode="val")

        # log step metric
        self.accuracy(preds, y)
        self.log("val_acc_step", self.accuracy)

    def test_step(self, batch, batch_idx):
        loss, preds, y = self.forward(batch, mode="test")

        # log step metric
        self.accuracy(preds, y)
        self.log("test_acc_step", self.accuracy)


def train_node_classifier(dataset, **model_kwargs):
    node_data_loader = DataLoader(dataset, batch_size=1)

    # Create a PyTorch Lightning trainer with the generation callback
    trainer = pl.Trainer(
        accelerator="gpu", devices=1, max_epochs=1000, enable_progress_bar=True
    )  # False because epoch size is 1

    # Note: the dimensions are specific for the Elliptic dataset
    model = GNN(**model_kwargs)
    trainer.fit(
        model, train_dataloaders=node_data_loader, val_dataloaders=node_data_loader
    )

    # Test best model on the test set
    trainer.test(model, node_data_loader, verbose=True)

    return model


node_gnn_model = train_node_classifier(
    dataset=elliptic_dataset, c_in=166, c_hidden=256, c_out=3, num_layers=4, dp_rate=0.1
)



import logging
from datetime import date
from pathlib import Path

from fintorch.datasets import stockticker

# Set logging level to INFO
logging.basicConfig(level=logging.INFO)

# Parameters
tickers = ["AAPL", "MSFT", "GOOG"]
data_path = Path("~/.fintorch_data/stocktickers/").expanduser()
start_date = date(2015, 1, 1)
end_date = date(2023, 6, 30)

# Create a dictionary mapping from tickers to index
ticker_index = {ticker: index for index, ticker in enumerate(tickers)}

# Load the stock dataset
stockdata = stockticker.StockTicker(
    data_path,
    tickers=tickers,
    start_date=start_date,
    end_date=end_date,
    mapping=ticker_index,
    force_reload=True,
)

print(stockdata.df_timeseries_dataset.to_pandas().describe())



from typing import Any
import lightning as L
import optuna
import torch
from optuna.integration import PyTorchLightningPruningCallback

# import wandb
from fintorch.datasets.ellipticpp import EllipticppDataModule
from fintorch.graph.layers.beanconv import BEANConvSimple
from fintorch.models.graph.graphbean.graphBEAN import GraphBEANModule

torch.set_float32_matmul_precision("medium")


def objective(trial: optuna.trial.Trial, max_epochs: int, predict: Any) -> float:
    # We optimize the number of layers, hidden units in each layer and dropouts.
    encoder_layers = trial.suggest_int("encoder_layers", 1, 4)
    decoder_layers = trial.suggest_int("decoder_layers", 1, 4)
    class_head_layers = trial.suggest_int("class_head_layers", 1, 10)
    hidden_layers = trial.suggest_int("hidden_layers", 8, 128)
    learning_rate = trial.suggest_float("learning_rate", 0.0001, 0.005)
    structure_decoder_head_layers = trial.suggest_int(
        "structure_decoder_head_layers", 2, 16
    )
    structure_decoder_head_out_channel = trial.suggest_int(
        "structure_decoder_head_out_channel", 8, 64
    )

    if predict == "transactions":
        edge = ("wallets", "to", "transactions")
    else:
        edge = ("transactions", "to", "wallets")

    model = GraphBEANModule(  # type: ignore[call-arg]
        edge,
        edge_types=[("wallets_to_transactions"), ("transactions_to_wallets")],
        encoder_layers=encoder_layers,
        decoder_layers=decoder_layers,
        learning_rate=learning_rate,
        class_head_layers=class_head_layers,
        hidden_layers=hidden_layers,
        structure_decoder_head_layers=structure_decoder_head_layers,
        structure_decoder_head_out_channels=structure_decoder_head_out_channel,
        conv_type=BEANConvSimple,
        classifier=True,
        predict=predict,
    )
    datamodule = EllipticppDataModule(edge, batch_size=2056)

    trainer = L.Trainer(
        enable_checkpointing=False,
        max_epochs=50,
        accelerator="auto",
        devices=1,
        logger=L.pytorch.loggers.WandbLogger(project=f"graphbean-sweep-{predict}"),
        callbacks=[PyTorchLightningPruningCallback(trial, monitor="val_acc_step")],
    )
    hyperparameters = dict(
        encoder_layers=encoder_layers,
        decoder_layers=decoder_layers,
        learning_rate=learning_rate,
        class_head_layers=class_head_layers,
        hidden_layers=hidden_layers,
    )
    trainer.logger.log_hyperparams(hyperparameters)  # type: ignore
    trainer.fit(model, datamodule=datamodule)

    return trainer.callback_metrics["val_f1"].item()



import os
import tempfile
import unittest
from datetime import datetime

import polars as pol
import pytest

from fintorch.datasets.marketdata import MarketDataset


class TestMarketDataset(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for testing
        self.test_dir = tempfile.mkdtemp()
        self.split = "train"

        # Mock raw data for testing
        self.raw_dir = os.path.join(self.test_dir, "raw")
        os.makedirs(self.raw_dir, exist_ok=True)
        raw_data = pol.DataFrame(
            {
                "date_id": [0, 1],
                "time_id": [0, 1],
                "symbol_id": [101, 102],
                "responder_6": [0.5, 0.6],
                "feature_1": [1.0, 2.0],
                "feature_2": [3.0, 4.0],
            }
        )
        raw_data.write_parquet(os.path.join(self.raw_dir, f"{self.split}.parquet"))

    @pytest.mark.special
    def test_setup_directories(self):
        """Test that directories are set up correctly."""
        dataset = MarketDataset(root=self.test_dir, split=self.split)

        self.assertTrue(os.path.exists(self.test_dir))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "raw")))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "processed")))

        self.assertTrue(hasattr(dataset, "data"))
        # Collect the data to check its content
        collected_data = dataset.data.collect()
        self.assertFalse(collected_data.is_empty())
        self.assertIn("unique_id", collected_data.columns)
        self.assertIn("ds", collected_data.columns)

        expected_paths = [
            os.path.join(self.test_dir, "raw", filename)
            for filename in [
                "features.csv",
                "responders.csv",
                "sample_submission.csv",
                "train.parquet",
                "test.parquet",
                "lags.parquet",
            ]
        ]

        # Check that raw_paths returns the correct file paths
        self.assertEqual(dataset.raw_paths(), expected_paths)

        # Check that each file exists
        for path in expected_paths:
            self.assertTrue(
                os.path.isfile(path), f"Expected file does not exist: {path}"
            )

        # Check that the train folder exists in the processed folder
        processed_train_folder = os.path.join(self.test_dir, "processed", "train")
        self.assertTrue(
            os.path.exists(processed_train_folder)
            and os.path.isdir(processed_train_folder),
            f"Processed train folder does not exist or is not a directory: {processed_train_folder}",
        )

    def test_map_to_datetime(self):
        """Test the map_to_datetime method."""
        batch = pol.DataFrame(
            {
                "date_id": [0],
                "time_id": [0],
            }
        )
        result = MarketDataset.map_to_datetime(batch)
        self.assertIn("ds", result.columns)
        expected_datetime = datetime(2023, 1, 1, 12, 0, 0)
        self.assertEqual(result["ds"][0], expected_datetime)

    @pytest.mark.special
    def test_setup_directories_failure(self):
        """Test that setup_directories raises an exception if it fails."""
        with unittest.mock.patch(
            "os.makedirs", side_effect=OSError("Permission denied")
        ):
            with self.assertRaises(RuntimeError):
                dataset = MarketDataset(root=self.test_dir)
                dataset.setup_directories()

    def test_download_without_credentials(self):
        """Test the download method handles missing Kaggle credentials."""
        # Temporarily rename the Kaggle config file if it exists
        kaggle_config = os.path.expanduser("~/.kaggle/kaggle.json")
        if os.path.exists(kaggle_config):
            temp_config = kaggle_config + ".temp"
            os.rename(kaggle_config, temp_config)
        else:
            temp_config = None
        try:
            with self.assertRaises(Exception):
                MarketDataset(root=self.test_dir)
        finally:
            # Restore the Kaggle config file
            if temp_config:
                os.rename(temp_config, kaggle_config)



import torch
from torch_geometric.nn import SAGEConv, to_hetero

from fintorch.datasets import ellipticpp

# Load the elliptic dataset
actor_transaction_graph = ellipticpp.TransactionActorDataset("~/.fintorch_data")

data = actor_transaction_graph[0]


class GNN(torch.nn.Module):
    def __init__(self, hidden_channels, out_channels):
        super().__init__()
        self.conv1 = SAGEConv((-1, -1), hidden_channels)
        self.conv2 = SAGEConv((-1, -1), out_channels)

    def forward(self, x, edge_index):
        x = self.conv1(x, edge_index)
        x = self.conv2(x, edge_index)
        return x


model = GNN(hidden_channels=64, out_channels=3)
model = to_hetero(model, data.metadata(), aggr="sum")

with torch.no_grad():
    out = model(data.x_dict, data.edge_index_dict)
    print(out)



import csv
import logging

import torch
from neuralforecast import NeuralForecast
from neuralforecast.auto import AutoNBEATS, AutoNHITS, AutoTFT
from neuralforecast.losses.numpy import mae, mse
from neuralforecast.losses.pytorch import MAE, HuberLoss
from ray import tune
from ray.air.integrations.wandb import WandbLoggerCallback

from fintorch.datasets.marketdata import MarketDataset

# Set up logging
logging.basicConfig(level=logging.INFO)
torch.set_float32_matmul_precision("medium")

input_size = 2  # Number of past time steps used for prediction
days = 1  # Number of days to forecast
steps_per_day = 5  # Number of steps per day
horizon = days * steps_per_day  # Forecast horizon
max_steps = 1000  # Max training steps

# Define validation and test sizes
val_size = horizon  # Validation set size
test_size = horizon  # Test set size

num_samples = 10  # Number of samples for hyperparameter tuning

windows_batch_size = 16
valid_batch_size = 16
batch_size = 1

# Define configurations for Auto models
tft_config = {
    "input_size": tune.choice([2, 4, 6]),
    "hidden_size": tune.choice([64, 128, 256]),
    "learning_rate": tune.loguniform(1e-4, 1e-1),
    "scaler_type": tune.choice(["robust", "standard"]),
    "max_steps": max_steps,
    "batch_size": batch_size,
    "windows_batch_size": windows_batch_size,
    "valid_batch_size": valid_batch_size,
    "random_seed": tune.randint(1, 20),
    "futr_exog_list": ["feature_00", "feature_01", "feature_02"],
}

nbeats_config = {
    "input_size": tune.choice([2, 4, 6]),
    "learning_rate": tune.loguniform(1e-4, 1e-1),
    "max_steps": max_steps,
    "batch_size": batch_size,
    "valid_batch_size": valid_batch_size,
    "windows_batch_size": windows_batch_size,
    "random_seed": tune.randint(1, 20),
}

nhits_config = {
    "input_size": tune.choice([2, 4, 6]),
    "learning_rate": tune.loguniform(1e-4, 1e-1),
    "max_steps": max_steps,
    "batch_size": batch_size,
    "valid_batch_size": valid_batch_size,
    "windows_batch_size": windows_batch_size,
    "random_seed": tune.randint(1, 20),
    "futr_exog_list": ["feature_00", "feature_01", "feature_02"],
}

# Set up WandbLoggerCallback
wandb_callback_tft = WandbLoggerCallback(
    project="jane-tft",  # Replace with your W&B project name
    entity="ai4fintech",  # Replace with your W&B username or team name
    log_config=True,
)

wandb_callback_nbeats = WandbLoggerCallback(
    project="jane-nbeats",  # Replace with your W&B project name
    entity="ai4fintech",  # Replace with your W&B username or team name
    log_config=True,
)

wandb_callback_nhits = WandbLoggerCallback(
    project="jane-nhits",  # Replace with your W&B project name
    entity="ai4fintech",  # Replace with your W&B username or team name
    log_config=True,
)

# Define ray_tune_kwargs with the Wandb callback
ray_tune_kwargs_tft = [wandb_callback_tft]
ray_tune_kwargs_nbeats = [wandb_callback_nbeats]
ray_tune_kwargs_nhits = [wandb_callback_nhits]

# Initialize the models
models = [
    AutoTFT(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=tft_config,
        num_samples=num_samples,
        refit_with_val=True,
        callbacks=ray_tune_kwargs_tft,
    ),
    AutoNBEATS(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=nbeats_config,
        num_samples=num_samples,
        refit_with_val=True,
        callbacks=ray_tune_kwargs_nbeats,
    ),
    AutoNHITS(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=nhits_config,
        num_samples=num_samples,
        refit_with_val=True,
        callbacks=ray_tune_kwargs_nhits,
    ),
]

# List of model names
model_names = ["AutoTFT", "AutoNBEATS", "AutoNHITS"]

# Create NeuralForecast object
nf = NeuralForecast(models=models, freq="10s")

# Load data
df = MarketDataset("~/.fintorch_data/marketdata-janestreet/")
df = df.data.collect()
# Number of unique series
n_series = len(df["unique_id"].unique())

# Perform cross-validation
Y_hat_df = nf.cross_validation(
    df=df,
    val_size=val_size,
    test_size=test_size,
    n_windows=None,  # Uses expanding window if None
).to_pandas()

# Iterate over models to compute metrics
results = []
for model_name in model_names:
    # Extract true values and predictions
    y_true = Y_hat_df.y.values
    y_hat = Y_hat_df[model_name].values

    # Reshape arrays
    y_true = y_true.reshape(n_series, -1, horizon)
    y_hat = y_hat.reshape(n_series, -1, horizon)

    # Compute metrics
    mse_score = mse(y_true, y_hat)
    mae_score = mae(y_true, y_hat)

    # Print results
    print(f"\nModel: {model_name}")
    print(f"MSE: {mse_score}")
    print(f"MAE: {mae_score}")

    # Append results to the list
    results.append({"Model": model_name, "MSE": mse_score, "MAE": mae_score})

# Save results to a CSV file
file_name = "model_scores.csv"
with open(file_name, "w", newline="") as file:
    writer = csv.DictWriter(file, fieldnames=["Model", "MSE", "MAE"])
    writer.writeheader()
    writer.writerows(results)



import lightning as L

from fintorch.datasets.ellipticpp import EllipticppDataModule
from fintorch.graph.layers.beanconv import BEANConvSimple
from fintorch.models.graph.graphbean.graphBEAN import GraphBEANModule


def main():
    # We use an example data module from the elliptic dataset which is bipartite
    data_module = EllipticppDataModule(
        ("wallets", "to", "transactions"), force_reload=False
    )
    try:
        data_module.setup()
        # Get the dimensionalities for the auto-encoder part
        mapping = dict()
        for key in data_module.dataset.metadata()[0]:
            mapping[key] = data_module.dataset[key].x.shape[1]
    except Exception as e:
        print(f"Error setting up data module or retrieving metadata: {e}")
        raise

    # Create an instance of the GraphBEANModule
    module = GraphBEANModule(
        ("wallets", "to", "transactions"),
        edge_types=[
            ("wallets", "to", "transactions"),
            ("transactions", "to", "wallets"),
        ],
        mapping=mapping,
        learning_rate=0.001,
        conv_type=BEANConvSimple,
        encoder_layers=5,
        decoder_layers=5,
        hidden_layers=50,
    )

    # Create a PyTorch Lightning Trainer and train the module
    trainer = L.Trainer(max_epochs=100, accelerator="auto")

    # Train the module using the dataloaders
    trainer.fit(module, datamodule=data_module)


if __name__ == "__main__":
    main()



import logging
import os
from datetime import date as Date
from typing import Any, Dict, List

import pandas as pd
import polars as pol
import yfinance as yf  # type: ignore
from neuralforecast.tsdataset import TimeSeriesDataset  # type: ignore
from tenacity import retry, stop_after_attempt, wait_exponential  # type: ignore
from torch.utils.data import Dataset


# TODO: directly subclass TimeSeriesDataset from neural forcast
class StockTicker(Dataset):  # type: ignore
    def __init__(
        self,
        root: str,
        tickers: List[str],
        start_date: Date,
        end_date: Date,
        mapping: Dict[str, str],  # TODO: check dict type
        value_name: str = "Adj Close",
        force_reload: bool = False,
    ) -> None:
        super().__init__()
        assert isinstance(tickers, list), "tickers must be a list"
        assert isinstance(start_date, Date), "start_date must be a Date object"
        assert isinstance(end_date, Date), "end_date must be a Date object"

        # Check if the start_date is before the end_date
        if start_date >= end_date:
            raise ValueError("start_date must be before end_date.")

        self.tickers = tickers
        self.start_date = start_date
        self.end_date = end_date
        self.value_name = value_name
        self.mapping = mapping
        self.root = os.path.expanduser(root)

        logging.info("StockTicker dataset initialization")
        self.setupDirectories()

        if force_reload or not all(
            os.path.exists(path) for path in self.processed_paths()
        ):
            # if we want to force reload, or a processed file is missing. Start the processing
            self.download(force_reload)  # download uncached tickers
            self.process()

        self.load()

    def setupDirectories(self) -> None:
        """
        Sets up the necessary directories for storing raw and processed data.

        This method checks if the root directory exists, and if not, creates it.
        It also checks if the raw and processed directories exist within the root directory,
        and if not, creates them.

        Args:
            None

        Returns:
            None
        """

        try:
            os.makedirs(self.root, exist_ok=True)
            os.makedirs(os.path.join(self.root, "raw"), exist_ok=True)
            os.makedirs(os.path.join(self.root, "processed"), exist_ok=True)
        except OSError as e:
            logging.error(f"Failed to create directories: {str(e)}")
            raise RuntimeError(f"Failed to setup directories: {str(e)}")

    def __len__(self) -> int:
        """Return the length of the dataset."""
        if not hasattr(self, "timeseries_dataset"):
            raise RuntimeError("Dataset not loaded. Call load() first.")
        return len(self.timeseries_dataset)

    def __getitem__(self, idx: int) -> Any:
        """
        Get an item from the dataset

        Args:
            idx: Index of the item to retrieve

        Returns:
            TimeseriesDataset: A slice of stock tick data with spatial graph encoding

        Raises:
            RuntimeError: If the dataset hasn't been loaded
        """
        if not hasattr(self, "timeseries_dataset"):
            raise RuntimeError("Dataset not loaded. Call load() first.")
        return self.timeseries_dataset.__getitem__(idx)

    def raw_file_names(self) -> list[str]:
        """
        Generates a list of raw file names based on the tickers, start date, and end date.

        Returns:
            A list of raw file names in the format "{ticker}_date_range_start_{start_date}_end_{end_date}.csv".
        """
        return [
            f"{ticker}_date_range_start_{self.start_date}_end_{self.end_date}.csv"
            for ticker in self.tickers
        ]

    def raw_paths(self) -> list[str]:
        # Return the full paths to the raw files
        # If you don't need to download raw files, return an empty list
        return [os.path.join(self.root, "raw", name) for name in self.raw_file_names()]

    def processed_paths(self) -> list[str]:
        """
        Returns a list of full paths to the processed files.

        Returns:
            list[str]: A list of full paths to the processed files.
        """
        return [
            os.path.join(self.root, "processed", name)
            for name in self.processed_file_names()
        ]

    def processed_file_names(self) -> list[str]:
        """
        Returns a list of processed file names.

        Returns:
            list[str]: A list of processed file names.
        """
        return [
            "timeseries_stocks_v1_partitioned",
        ]

    @retry(  # type: ignore[misc]
        stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10)
    )
    def download_with_retry(
        self, tickers: List[str], start_date: Date, end_date: Date
    ) -> Any:
        return yf.download(tickers, start=start_date, end=end_date)

    def download(self, force_reload: bool = False) -> None:
        """
        Downloads the raw stock data from Yahoo Finance for the specified tickers and time range.

        Args:
            force_reload (bool, optional): If True, forces the download even if the raw data files already exist.
                Defaults to False.

        Raises:
            ValueError: If the specified value_name does not exist as a column in the raw_data dataframe.
        """
        if not all(os.path.exists(path) for path in self.raw_paths()) or force_reload:
            # Download the raw data for tickers that don't have the raw files
            if force_reload:
                missing_tickers = self.tickers
                logging.info("force reloading stock data from yahoo finance")
            else:
                missing_tickers = [
                    ticker
                    for ticker, path in zip(self.tickers, self.raw_paths())
                    if not os.path.exists(path)
                ]
                logging.info(f"Only downloading missing tickers: {missing_tickers}")

            raw_data = self.download_with_retry(
                missing_tickers, self.start_date, self.end_date
            )

            # Check if self.value_name exists as a column in the raw_data dataframe
            if self.value_name not in raw_data.columns:
                raise ValueError(
                    f"{self.value_name} does not exist as a column in the raw_data dataframe"
                )

            # Reshape the data
            df = raw_data[self.value_name]  # No need to unstack here

            # Convert the Series to a DataFrame if it's not already (optional but recommended)
            if isinstance(df, pd.Series):
                df = df.to_frame()

            # Melt the dataframe to long format
            hist = df.melt(
                ignore_index=False, var_name="Ticker", value_name=self.value_name
            )
            hist.reset_index(inplace=True)

            hist.rename(
                columns={"Date": "ds", "Ticker": "unique_id", self.value_name: "y"},
                inplace=True,
            )

            # Split the hist dataframe based on unique_id
            grouped_data = hist.groupby("unique_id")

            # Save a csv file per unique_id
            for ticker, group in grouped_data:
                file_name = f"{ticker}_date_range_start_{self.start_date}_end_{self.end_date}.csv"
                group.to_csv(os.path.join(self.root, "raw", file_name), index=False)
        else:
            logging.info(
                "Using cached data, if you want to force reload set force_reload=True"
            )

    def process(self) -> None:
        """
        Process the stock ticker data.

        Reads the CSV files provided by `raw_files` using polars library.
        Concatenates the files into one data frame.
        Reshapes the data to wide format based on unique_id.
        Constructs a spatial graph based on the wide data frame.
        Stores the time-series data in a parquet file.
        Stores the spatial graph using torch.save.

        Returns:
            None
        """
        # Read all the csv files provided by raw_files with polars
        dfs = []
        for file_path in self.raw_paths():
            try:
                df = pol.read_csv(file_path)
                if df.is_empty():
                    raise ValueError(f"Empty data in file {file_path}")
            except Exception as e:
                logging.error(f"Failed to process {file_path}: {str(e)}")
                raise

            dfs.append(df)

        # Concatenate the files into one data frame
        concatenated_df = pol.concat(dfs)
        concatenated_df = concatenated_df.with_columns(
            pol.col("unique_id").replace(self.mapping, default=None),
            pol.col("ds").str.to_date(),
        )

        # Reshape the data to wide format based on unique_id
        wide_df = concatenated_df.pivot(index="ds", on="unique_id", values="y")
        wide_df = wide_df.drop(["ds"])

        # Store time-series data
        concatenated_df.write_parquet(
            self.processed_paths()[0],
            compression="zstd",
            use_pyarrow=True,
            pyarrow_options={"partition_cols": ["unique_id"]},
        )

    def load(self) -> None:
        """
        Loads the dataset and associated graphs.

        This method loads the dataset from a parquet file and creates the necessary graphs for further processing.
        It also logs the loaded dataset, spatial graph, and temporal encoding.

        Returns:
            None
        """

        try:
            self.df_timeseries_dataset = pol.read_parquet(self.processed_paths()[0])

            self.timeseries_dataset, _, _, _ = TimeSeriesDataset.from_df(
                self.df_timeseries_dataset,
                id_col="unique_id",
                time_col="ds",
                target_col="y",
            )
        except Exception as e:
            logging.error(f"Failed to load dataset: {str(e)}")
            raise RuntimeError("Failed to load dataset") from e
        logging.info("All datsets loaded sucessfully")

        logging.info(f"loaded dataset:{self.timeseries_dataset}")



import logging
from pathlib import Path

from fintorch.datasets.marketdata import MarketDataset
from torch.utils.data import DataLoader

logging.basicConfig(level=logging.INFO)

data_path = Path("~/.fintorch_data/marketdata-janestreet/").expanduser()


market_data = MarketDataset(
    root=data_path, force_reload=False, split="train", batch_size=1024
)
# Batch size is determined by the Market_data
dataloader = DataLoader(market_data, batch_size=None, num_workers=28)

for _, batch in dataloader:
    print("Loading train batch")
    print(batch.describe())

market_data = MarketDataset(root=data_path, force_reload=False, split="test")
# Batch size is determined by the Market_data
dataloader = DataLoader(market_data, batch_size=None, num_workers=10)

for _, batch in dataloader:
    print("Loading test batch")
    print(batch.describe())



from click.testing import CliRunner

from fintorch import cli
from fintorch.datasets import elliptic, ellipticpp


def test_command_line_interface():
    runner = CliRunner()
    result = runner.invoke(cli.fintorch)
    assert result.exit_code == 0
    assert "FinTorch CLI - Your financial AI toolkit" in result.output
    help_result = runner.invoke(cli.fintorch, ["--help"])
    assert help_result.exit_code == 0
    assert "--help  Show this message and exit." in help_result.output


def test_command_line_interface_elliptic_dataset():
    runner = CliRunner()
    result = runner.invoke(cli.fintorch, ["datasets", "elliptic"])
    assert result.exit_code == 0
    assert "Downloading dataset: elliptic" in result.output
    root = "/tmp/data/fintorch/"
    assert isinstance(
        elliptic.TransactionDataset(root, force_reload=True),
        elliptic.TransactionDataset,
    )


def test_command_line_interface_ellipticpp_dataset():
    runner = CliRunner()
    result = runner.invoke(cli.fintorch, ["datasets", "ellipticpp"])
    assert result.exit_code == 0
    assert "Downloading dataset: ellipticpp" in result.output
    root = "/tmp/data/fintorch/"
    assert isinstance(
        ellipticpp.TransactionActorDataset(root, force_reload=True),
        ellipticpp.TransactionActorDataset,
    )


def test_command_line_interface_train():
    runner = CliRunner()
    result = runner.invoke(
        cli.fintorch, ["sweep", "--model", "model_name", "--predict", "predict"]
    )
    assert "Starting sweep for model" in result.output



import os
import shutil
import tempfile
import unittest

import polars as pol
import torch

from fintorch.datasets.auctiondata import AuctionDataset


class TestAuctionDataset(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for testing
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the temporary directory after tests
        shutil.rmtree(self.test_dir)

    def test_setupDirectories(self):
        """Test that directories are set up correctly."""
        AuctionDataset(root=self.test_dir)
        self.assertTrue(os.path.exists(self.test_dir))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "raw")))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "processed")))

    def test_processed_paths(self):
        """Test that processed_paths returns the correct file paths."""
        dataset = AuctionDataset(root=self.test_dir)
        expected_paths = [
            os.path.join(self.test_dir, "raw/train.csv"),
            os.path.join(self.test_dir, "raw/example_test_files/revealed_targets.csv"),
            os.path.join(self.test_dir, "raw/example_test_files/sample_submission.csv"),
            os.path.join(self.test_dir, "raw/example_test_files/test.csv"),
        ]
        self.assertEqual(dataset.processed_paths(), expected_paths)

    def test_load(self):
        dataset = AuctionDataset(root=self.test_dir)
        dataset.load()

        # Assertions for train DataFrame
        self.assertTrue(hasattr(dataset, "train"))
        self.assertIn("y", dataset.train.columns)
        self.assertIn("unique_id", dataset.train.columns)
        self.assertNotIn("row_id", dataset.train.columns)
        self.assertNotIn("near_price", dataset.train.columns)
        self.assertNotIn("far_price", dataset.train.columns)

        # Assertions for test DataFrame
        self.assertTrue(hasattr(dataset, "test"))
        self.assertIn("unique_id", dataset.test.columns)
        self.assertNotIn("near_price", dataset.test.columns)
        self.assertNotIn("far_price", dataset.test.columns)

    def test_len(self):
        """Test that __len__ returns the correct length."""
        dataset = AuctionDataset(root=self.test_dir)
        # Mock the train DataFrame
        dataset.train = pol.DataFrame({"col1": [1, 2, 3]})
        self.assertEqual(len(dataset), 3)

    def test_getitem(self):
        """Test that __getitem__ retrieves data correctly."""
        dataset = AuctionDataset(root=self.test_dir)
        # Mock the train DataFrame
        dataset.train = pol.DataFrame(
            {"col1": [1, 2, 3], "col2": [4, 5, 6], "ds": [7, 8, 9]}
        )
        item = dataset[1]
        expected_tensor = torch.tensor([2, 5], dtype=torch.float32)
        self.assertTrue(torch.equal(item, expected_tensor))

    def test_process(self):
        """Test the process method (currently a placeholder)."""
        dataset = AuctionDataset(root=self.test_dir)
        try:
            dataset.process()
            self.assertTrue(True)  # Method executed without error
        except Exception as e:
            self.fail(f"Process method raised an exception: {e}")



import lightning as L
import torch_geometric.transforms as T
from torch_geometric.datasets import OGB_MAG
from torch_geometric.loader import LinkNeighborLoader

from fintorch.graph.layers.beanconv import BEANConvSimple
from fintorch.models.graph.graphbean.graphBEAN import GraphBEANModule


def main():
    class OGBDataModule(L.LightningDataModule):
        def __init__(self, data, edge):
            super().__init__()
            self.dataset = data
            self.edge = edge

        def train_dataloader(self):
            # Define neighbor sampling configuration
            src, to, dst = self.edge
            num_neighbors = {
                ("paper", "cites", "paper"): [10, 10],
                ("author", "writes", "paper"): [10, 10],
                ("paper", "has_topic", "field_of_study"): [10, 10],
                ("author", "affiliated_with", "institution"): [10, 10],
                ("institution", "rev_affiliated_with", "author"): [10, 10],
                ("paper", "rev_writes", "author"): [10, 10],
                ("field_of_study", "rev_has_topic", "paper"): [10, 10],
            }
            loader = LinkNeighborLoader(
                self.dataset,
                num_neighbors=num_neighbors,
                batch_size=512,
                shuffle=True,
                edge_label_index=(
                    (src, to, dst),
                    self.dataset[src, to, dst].edge_index,
                ),
                neg_sampling="binary",
                num_workers=14,
            )

            return loader

        def val_dataloader(self):
            # Define neighbor sampling configuration
            src, to, dst = self.edge
            num_neighbors = {
                ("paper", "cites", "paper"): [10, 10],
                ("author", "writes", "paper"): [10, 10],
                ("paper", "has_topic", "field_of_study"): [10, 10],
                ("author", "affiliated_with", "institution"): [10, 10],
                ("institution", "rev_affiliated_with", "author"): [10, 10],
                ("paper", "rev_writes", "author"): [10, 10],
                ("field_of_study", "rev_has_topic", "paper"): [10, 10],
            }
            loader = LinkNeighborLoader(
                self.dataset,
                num_neighbors=num_neighbors,
                batch_size=128,
                edge_label_index=(
                    (src, to, dst),
                    self.dataset[src, to, dst].edge_index,
                ),
                neg_sampling="binary",
                num_workers=14,
            )

            return loader

    dataset = OGB_MAG(root="./data", preprocess="metapath2vec")
    data = T.ToUndirected()(dataset[0])
    data_module = OGBDataModule(data, ("author", "writes", "paper"))

    try:
        # Get the dimensionalities for the auto-encoder part
        mapping = dict()
        for key in data_module.dataset.metadata()[0]:
            mapping[key] = data_module.dataset[key].x.shape[1]
    except Exception as e:
        print(f"Error retrieving metadata: {e}")
        raise
    # # Create an instance of the GraphBEANModule
    module = GraphBEANModule(
        ("author", "writes", "paper"),
        edge_types=[
            ("author", "writes", "paper"),
            ("paper", "rev_writes", "author"),
            ("author", "affiliated_with", "institution"),
            ("institution", "rev_affiliated_with", "author"),
        ],
        mapping=mapping,
        learning_rate=0.001,
        conv_type=BEANConvSimple,
        encoder_layers=5,
        decoder_layers=5,
        hidden_layers=50,
        predict="paper",
    )

    # Create a PyTorch Lightning Trainer and train the module
    trainer = L.Trainer(max_epochs=100, accelerator="auto")

    trainer.fit(module, datamodule=data_module)


if __name__ == "__main__":
    main()



import logging
import os
from datetime import datetime
from typing import Any, List
from zipfile import ZipFile

import polars as pol
import torch
from torch.utils.data import Dataset


class AuctionDataset(Dataset):  # type: ignore
    """
    AuctionDataset Class Documentation

    The `AuctionDataset` class provides a convenient wrapper around the auction dataset used for the NASDAQ daily ten-minute closing auction for the Kaggle competition trading at the close from Optiver. It allows for the automated downloading, processing, and loading of data, enabling efficient handling and use within machine learning workflows, specifically those using PyTorch's Dataset utilities.

    Attributes:
    root (str): The root directory where the dataset will be downloaded, stored, and processed.
    force\_reload (bool): If True, forces a redownload and reprocessing of the dataset even if the required files are already present.

    Methods:
    setupDirectories():
    Creates the necessary directory structure for storing raw and processed data.
    download():
    Downloads the dataset from Kaggle using the Kaggle API.
    process():
    Placeholder for applying transformations to the downloaded data to prepare it for modeling.
    load():
    Loads the processed data for easy access.
    **len**():
    Returns the length of the dataset.
    **getitem**(idx):
    Allows indexing and iteration over data samples.

    Overview:
    The `AuctionDataset` class is designed to load and preprocess historical auction data provided by the Optiver competition on Kaggle. This data is used to predict future price movements of NASDAQ-listed stocks relative to a synthetic index. The dataset includes time series data with multiple features such as stock price, bid/ask information, imbalance data, and reference prices.

    Key Features:
    - Automated Data Download: Automatically downloads the dataset from Kaggle using the Kaggle API. The dataset is sourced from the Optiver "Trading at the Close" competition.
    - Directory Setup: Manages directories for storing both raw and processed data, making it straightforward to maintain organized data structures.
    - Processing Workflow: Provides an interface to transform the raw auction data as needed. This is useful for creating features suitable for training predictive models.
    - Seamless Integration: Extends `torch.utils.data.Dataset`, making it compatible with PyTorch's data loaders and facilitating its integration into machine learning pipelines.

    Usage:
    The class handles the entire lifecycle of the dataset:

    ```
    - Downloading: When instantiated, if required files are missing or if forced reload is requested, the dataset is downloaded directly from Kaggle.
    - Processing: After downloading, the data is processed to apply transformations that make it suitable for model training and evaluation.
    - Loading: Finally, the processed data is loaded into memory, making it available for access by PyTorch.
    ```

    Constructor Parameters:
    root (str): The root directory where the dataset will be downloaded, stored, and processed.
    force_reload (bool, default=False): If True, forces a redownload and reprocessing of the dataset even if the required files are already present.

    Methods Overview:
    setupDirectories():
    Creates the necessary directory structure for storing raw and processed data. Ensures that the root, raw, and processed directories exist.

    ```
    download():
        Uses the Kaggle API to download the dataset. It reads configuration details for authenticating and retrieving files from the Optiver competition.

    process():
        Placeholder function where transformations are applied to the downloaded data to prepare it for modeling.

    load():
        Loads the processed data for easy access when training or evaluating machine learning models.

    __len__() and __getitem__(idx):
        Implementations required by `torch.utils.data.Dataset` to allow indexing and iteration over data samples.
    ```

    Practical Application:
    This class is useful in financial data analysis, specifically for training predictive models using auction data from NASDAQ. It abstracts the tedious parts of dataset handling—such as downloading, file management, and directory creation—allowing data scientists and researchers to focus more on model building and experimentation.


    Train and Test Sets:
    The dataset is split into a training set (`train.csv`) and a test set  (test.csv) . The training set contains historical auction data used for model training, while the test set is used to evaluate model performance. The test set is provided through the competition API and contains similar features but without target labels.

    Competition Data:
    In this competition, Kagglers are challenged to predict the short-term price movements during the 10-minute auction period. All the training data is contained within a single `train.csv` file. Please refer to [https://www.kaggle.com/competitions/optiver-trading-at-the-close/data](https://www.kaggle.com/competitions/optiver-trading-at-the-close/data) & the introduction in the associated notebook for specific column definitions.

    ```
    The dataset includes the following columns:
    - `stock_id`: A unique identifier for the stock. Not all stock IDs exist in every time bucket.
    - `date_id`: A unique identifier for the date. Date IDs are sequential and consistent across all stocks.
    - `imbalance_size`: The amount unmatched at the current reference price (in USD).
    - `imbalance_buy_sell_flag`: An indicator reflecting the direction of auction imbalance.
        - Buy-side imbalance: `1`
        - Sell-side imbalance: `-1`
        - No imbalance: `0`
    - `reference_price`: The price at which paired shares are maximized, the imbalance is minimized, and the distance from the bid-ask midpoint is minimized, in that order. Can also be thought of as equal to the near price bounded between the best bid and ask price.
    - `matched_size`: The amount that can be matched at the current reference price (in USD).
    - `far_price`: The crossing price that will maximize the number of shares matched based on auction interest only, excluding continuous market orders.
    - `near_price`: The crossing price that will maximize the number of shares matched based on auction and continuous market orders.
    - `[bid/ask]_price`: Price of the most competitive buy/sell level in the non-auction book.
    - `[bid/ask]_size`: The dollar notional amount on the most competitive buy/sell level in the non-auction book.
    - `wap`: The weighted average price in the non-auction book, calculated as:

    (BidPrice * AskSize + AskPrice * BidSize) / (BidSize + AskSize)

    - `seconds_in_bucket`: The number of seconds elapsed since the beginning of the day's closing auction, always starting from 0.
    - `target`: The 60-second future move in the `wap` of the stock, less the 60-second future move of the synthetic index. The synthetic index is a custom weighted index of Nasdaq-listed stocks constructed by Optiver for this competition. The unit of the target is basis points, where a 1 basis point price move is equivalent to a 0.01% price move.

    Where `t` is the time at the current observation, we can define the target as:

    Target = ((StockWAP[t+60] / StockWAP[t]) - (IndexWAP[t+60] / IndexWAP[t])) * 10000

    All size-related columns are in USD terms.
    All price-related columns are converted to a price move relative to the stock `wap` (weighted average price) at the beginning of the auction period.
    ```

    Citation:
    Tom Forbes, John Macgillivray, Matteo Pietrobon, Sohier Dane, and Maggie Demkin. Optiver - Trading at the Close. [https://kaggle.com/competitions/optiver-trading-at-the-close](https://kaggle.com/competitions/optiver-trading-at-the-close), 2023. Kaggle.



    """

    def __init__(self, root: str, force_reload: bool = False):
        super().__init__()

        self.root = root

        logging.info("Load auction data")
        self.setupDirectories()

        if force_reload or not all(
            os.path.exists(path) for path in self.processed_paths()
        ):
            # if we want to force reload, or a processed file is missing. Start the processing
            self.download()  # download auction data
            self.process()

        self.load()

    def __len__(self) -> int:
        return self.train.shape[0]

    def __getitem__(self, idx: int) -> Any:
        return torch.tensor(self.train.drop("ds").row(idx), dtype=torch.float32)

    def processed_paths(self) -> List[str]:
        return [
            os.path.join(self.root, path)
            for path in [
                "raw/train.csv",
                "raw/example_test_files/revealed_targets.csv",
                "raw/example_test_files/sample_submission.csv",
                "raw/example_test_files/test.csv",
            ]
        ]

    def process(self) -> None:
        logging.info("Processing: apply transformations to auction data")

    def download(self) -> None:
        from kaggle.api.kaggle_api_extended import KaggleApi  # type: ignore

        logging.info("Downloading dataset from Kaggle")
        api = KaggleApi()
        api.read_config_file()
        api.authenticate()
        competition = "optiver-trading-at-the-close"
        api.competition_download_files(
            competition, path=os.path.join(self.root, "raw"), quiet=False
        )

        path = os.path.join(self.root, "raw", competition + ".zip")
        with ZipFile(path) as zObject:
            zObject.extractall(os.path.join(self.root, "raw"))

        os.remove(path)

    def load(self) -> None:
        path_train = os.path.join(self.root, "raw", "train.csv")
        # TODO: Can we replace NaN with 0?
        self.train = pol.read_csv(path_train)

        # Rename to NeuralForecast convention
        self.train = self.train.rename({"target": "y", "stock_id": "unique_id"})

        # TODO: check exclusion, can we fill these columns?
        self.train = self.train.drop(["row_id", "near_price", "far_price"])

        # TODO: check fill strategy, do we have better strategies?
        self.train = self.train.fill_nan(0)
        self.train = self.train.fill_null(0)

        # filter overlap
        # BUG: the train and test set overlap, split properly for time-series analysis
        self.train = self.train.filter(~pol.col("date_id").is_in([478, 479, 480]))
        self.train = self.train.filter(~pol.col("seconds_in_bucket").is_in([540]))

        # map to proper ds
        self.train = self.map_to_datetime(self.train)

        path_test = os.path.join(self.root, "raw", "example_test_files", "test.csv")
        self.test = pol.read_csv(path_test)

        self.test = self.map_to_datetime(self.test)

        # Rename to NeuralForecast convention
        self.test = self.test.rename({"stock_id": "unique_id"})

        # TODO: check exclusion, can we fill these columns?
        self.test = self.test.drop(["near_price", "far_price"])

        # TODO: check fill strategy, do we have better strategies?
        self.test = self.test.fill_nan(0)
        self.test = self.test.fill_null(0)

    def map_to_datetime(self, df: pol.DataFrame) -> pol.DataFrame:
        start_date = datetime(2023, 1, 1)
        return df.with_columns(
            (
                pol.lit(start_date)
                + (pol.col("date_id") * pol.duration(minutes=9))
                + pol.duration(hours=12)
                + (pol.col("seconds_in_bucket") * pol.duration(seconds=1))
            ).alias("ds")
        )

    def setupDirectories(self) -> None:
        try:
            os.makedirs(self.root, exist_ok=True)
            os.makedirs(os.path.join(self.root, "raw"), exist_ok=True)
            os.makedirs(os.path.join(self.root, "processed"), exist_ok=True)
        except OSError as e:
            logging.error(f"Failed to create directories: {str(e)}")
            raise RuntimeError(f"Failed to setup directories: {str(e)}")



import csv
import logging
from pathlib import Path

import polars as pl
import torch
from neuralforecast import NeuralForecast
from neuralforecast.auto import AutoNBEATS, AutoNHITS, AutoTFT
from neuralforecast.losses.numpy import mae, mse
from neuralforecast.losses.pytorch import MAE, HuberLoss
from ray import tune

from fintorch.datasets.auctiondata import AuctionDataset

logging.basicConfig(level=logging.INFO)
torch.set_float32_matmul_precision("medium")

data_path = Path("~/.fintorch_data/auctiondata-optiver/").expanduser()
auction_data = AuctionDataset(data_path, False)


input_size = 30
days = 3
steps_per_day = 55
horizon = days * steps_per_day
batch_size = 16
num_samples = 10

max_steps = tune.choice([500, 1500, 2000])

tft_config = {
    "input_size": tune.choice([10, 20, 30]),
    "hidden_size": tune.choice([64, 128, 256]),
    "n_head": tune.choice([4, 8]),
    "learning_rate": tune.loguniform(1e-4, 1e-1),
    "scaler_type": tune.choice([None, "robust", "standard"]),
    "max_steps": max_steps,
    "batch_size": batch_size,
    "windows_batch_size": tune.choice([128, 256, 512, 1024]),
    "random_seed": tune.randint(1, 20),
}

nbeats_config = {
    "input_size": tune.choice([10, 20, 30]),
    "n_harmonics": tune.choice([2, 3, 4]),
    "n_polynomials": tune.choice([1, 2, 3]),
    "stack_types": tune.choice([["trend", "seasonality"]]),
    "n_blocks": tune.choice([[3, 3], [4, 4], [5, 5]]),
    "mlp_units": tune.choice(
        [
            [[1024, 2048], [2048, 2048]],
            [[2048, 4096], [4096, 4096]],
            [[2048, 4096], [4096, 8192]],
        ]
    ),
    "dropout_prob_theta": 0.0,
    "activation": tune.choice(["ReLU", "LeakyReLU"]),
    "shared_weights": tune.choice([True, False]),
    "valid_loss": None,
    "max_steps": max_steps,
    "learning_rate": tune.loguniform(1e-5, 1e-2),
    "num_lr_decays": tune.choice([2, 3, 4]),
    "early_stop_patience_steps": -1,
    "val_check_steps": 100,
    "batch_size": batch_size,
    "valid_batch_size": None,
    "windows_batch_size": 1024,
    "inference_windows_batch_size": -1,
    "start_padding_enabled": False,
    "step_size": 1,
    "scaler_type": "identity",
    "random_seed": 1,
    "num_workers_loader": 0,
}

nhits_config = {
    # "learning_rate": tune.choice([1e-3]),                                     # Initial Learning rate
    "learning_rate": tune.loguniform(1e-5, 5e-3),
    "max_steps": max_steps,  # Number of SGD steps
    "input_size": tune.choice([7 * horizon]),  # input_size = multiplier * horizon
    "batch_size": batch_size,  # Number of series in windows
    "windows_batch_size": tune.choice([256]),  # Number of windows in batch
    "n_pool_kernel_size": tune.choice([[2, 2, 2], [16, 8, 1]]),  # MaxPool's Kernelsize
    "n_freq_downsample": tune.choice(
        [[(96 * 7) // 2, 96 // 2, 1], [(24 * 7) // 2, 24 // 2, 1], [1, 1, 1]]
    ),  # Interpolation expressivity ratios
    "dropout_prob_theta": tune.choice([0.5]),  # Dropout regularization
    "activation": tune.choice(["ReLU"]),  # Type of non-linear activation
    "n_blocks": tune.choice([[1, 1, 1]]),  # Blocks per each 3 stacks
    "mlp_units": tune.choice(
        [[[512, 512], [512, 512], [512, 512]]]
    ),  # 2 512-Layers per block for each stack
    "interpolation_mode": tune.choice(["linear"]),  # Type of multi-step interpolation
    "val_check_steps": tune.choice([100]),  # Compute validation every 100 epochs
    "random_seed": tune.randint(1, 10),
    "futr_exog_list": [
        "wap",
        "bid_price",
        "ask_price",
    ],
}


models = [
    AutoTFT(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=tft_config,
        num_samples=num_samples,
        refit_with_val=True,
    ),
    AutoNBEATS(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=nbeats_config,
        num_samples=num_samples,
        refit_with_val=True,
    ),
    AutoNHITS(
        h=horizon,
        loss=HuberLoss(delta=0.5),
        valid_loss=MAE(),
        config=nhits_config,
        num_samples=num_samples,
        refit_with_val=True,
    ),
]

# Create a NeuralForecast object
nf = NeuralForecast(models=models, freq="10s")

# Define validation and test sizes
val_size = horizon  # Number of days for validation
test_size = horizon  # Number of days for testing

train_df = auction_data.train.select(
    [
        "y",
        "ds",
        "unique_id",
        "wap",
        "imbalance_size",
        "imbalance_buy_sell_flag",
        "reference_price",
        "matched_size",
        "bid_price",
        "ask_price",
        "ask_size",
    ]
)

# Perform cross-validation
Y_hat_df = nf.cross_validation(
    df=train_df,
    val_size=val_size,
    test_size=test_size,
    n_windows=None,  # Uses expanding window if None
).to_pandas()


# List of model names
model_names = ["AutoTFT", "AutoNBEATS", "AutoNHITS"]

# Number of unique series
n_series = len(auction_data.train["unique_id"].unique())

days_train = 479

# Initialize a list to store results
results = []

# Iterate over each model to compute and print MSE and MAE
for model_name in model_names:
    # Extract true values and predictions
    y_true = Y_hat_df.y.values
    y_hat = Y_hat_df[model_name].values

    # Reshape to (n_series, n_windows, horizon)
    y_true = y_true.reshape(n_series, -1, horizon)
    y_hat = y_hat.reshape(n_series, -1, horizon)

    # Compute MSE and MAE
    mse_score = mse(y_true, y_hat)
    mae_score = mae(y_true, y_hat)

    # Print results
    print(f"\nModel: {model_name}")
    print(f"MSE: {mse_score}")
    print(f"MAE: {mae_score}")

    # Append results to the list
    results.append({"Model": model_name, "MSE": mse_score, "MAE": mae_score})

# Save results to a CSV file
file_name = "model_scores.csv"
with open(file_name, "w", newline="") as file:
    writer = csv.DictWriter(file, fieldnames=["Model", "MSE", "MAE"])
    writer.writeheader()
    writer.writerows(results)

print(f"\nScores saved to {file_name}")


# Create the future dataframe
fcsts_df = nf.make_future_dataframe()

selected_data = auction_data.test.select(
    [
        "unique_id",
        "ds",
        "wap",
        "imbalance_size",
        "imbalance_buy_sell_flag",
        "reference_price",
        "matched_size",
        "bid_price",
        "ask_price",
        "ask_size",
        "row_id",
        "time_id",
    ]
)

# BUG: without y-values the code doesn't run, NeuralForecast requires a y column eventhough the documentation
#     states that it is not required for the futr_df column
selected_data = selected_data.with_columns(pl.lit(0).alias("y"))
fcsts_df = nf.predict(futr_df=selected_data)

print("Predictions on the test dataset by all models:")
fcsts_df = fcsts_df.join(selected_data, on=["unique_id", "ds"], how="right")
print(fcsts_df)
